// Copyright (c) 2016 Douglas Miller
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

public class H19Keyboard implements PasteListener, KeyListener, Runnable {
	OutputStream _fout;
	boolean _loopback;
	boolean insertToggle;
	Vector<ResetListener> resetees;
	AuxKeyListener auxkeyer = null;
	java.util.concurrent.LinkedBlockingDeque<String> fifo;
	int paste_delay = 33;	// mS, 1000/cps
	int cr_delay = 100;	// mS

	public H19Keyboard(OutputStream fout, boolean loopback) {
		_fout = fout;
		_loopback = loopback;
		insertToggle = false;
		resetees = new Vector<ResetListener>();
		fifo = new java.util.concurrent.LinkedBlockingDeque<String>();
		Thread t = new Thread(this);
		t.start();
	}

	public void sendBack(String s) {
		byte[] bs = s.getBytes();
		try {
			_fout.write(bs);
			_fout.flush();
		} catch (Exception ee) {
		}
	}

	public void addResetListener(ResetListener rl) {
		resetees.add(rl);
	}

	public void addAuxKeyListener(AuxKeyListener ak) {
		auxkeyer = ak;
	}

	public void setPasteRate(int cps, int cr) {
		if (cps >= 0) {
			if (cps < 1) {
				cps = 1;
			}
			int dly = 1000 / cps;
			if (dly <= 0) {
				dly = 1;
			}
			paste_delay = dly;
		}
		if (cr >= 0) {
			cr_delay = cr;
		}
	}

	public void paste(String s) {
		// TODO: what is the right translation here?
		if (_loopback) {
			fifo.add(s.replaceAll("\n", "\r\n"));
		} else {
			fifo.add(s.replaceAll("\n", "\r"));
		}
	}

        public void keyTyped(KeyEvent e) { }

        public void keyPressed(KeyEvent e) {
		String s = null;
                char c = e.getKeyChar();
                int k = e.getKeyCode();
  		int m = e.getModifiers();
		int l = e.getKeyLocation();
		if (l == KeyEvent.KEY_LOCATION_NUMPAD &&
				c != KeyEvent.CHAR_UNDEFINED &&
				(m & InputEvent.SHIFT_MASK) == 0) {
			if (auxkeyer != null) {
				auxkeyer.auxKeyPressed(e);
				return;
			}
		}
  		// Assume if CTRL is down, must be ^J not ENTER...
  		if (k == KeyEvent.VK_ENTER && (m & InputEvent.CTRL_MASK) == 0) {
  			s = "\015";
			if (_loopback) {
				s += "\012";
			}
  		}
		if (k == KeyEvent.VK_BACK_SPACE && (m & InputEvent.SHIFT_MASK) != 0) {
			c = 0x7f;
		}
		if (k == KeyEvent.VK_F1) {
			s = "\033S";
		} else if (k == KeyEvent.VK_F2) {
			s = "\033T";
		} else if (k == KeyEvent.VK_F3) {
			s = "\033U";
		} else if (k == KeyEvent.VK_F4) {
			s = "\033V";
		} else if (k == KeyEvent.VK_F5) {
			s = "\033W";
		} else if (k == KeyEvent.VK_F6) {
			s = "\033P";
		} else if (k == KeyEvent.VK_F7) {
			s = "\033Q";
		} else if (k == KeyEvent.VK_F8) {
			s = "\033R";
		} else if (k == KeyEvent.VK_UP || k == KeyEvent.VK_KP_UP) {
			s = "\033A";
		} else if (k == KeyEvent.VK_DOWN || k == KeyEvent.VK_KP_DOWN) {
			s = "\033B";
		} else if (k == KeyEvent.VK_LEFT || k == KeyEvent.VK_KP_LEFT) {
			s = "\033D";
		} else if (k == KeyEvent.VK_RIGHT || k == KeyEvent.VK_KP_RIGHT) {
			s = "\033C";
		} else if (k == KeyEvent.VK_HOME) {
			s = "\033H";
		} else if (k == KeyEvent.VK_DELETE) {
			if ((m & InputEvent.SHIFT_MASK) == 0) {
				s = "\033N";
			} else {
				s = "\033M";
			}
		} else if (k == KeyEvent.VK_INSERT) {
			if ((m & InputEvent.SHIFT_MASK) == 0) {
				if (auxkeyer != null) {
					auxkeyer.auxKeyPressed(e);
					return;
				}
			} else {
				s = "\033L";
			}
		} else if (k == KeyEvent.VK_SCROLL_LOCK) {
			if (auxkeyer != null) {
				auxkeyer.auxKeyPressed(e);
				return;
			}
		} else if (k == KeyEvent.VK_F9 &&
				(m & InputEvent.SHIFT_MASK) != 0) {
			for (int x = 0; x < resetees.size(); ++x) {
				resetees.get(x).resetPerformed();
			}
		}
		if (s != null) {
			e.consume();
			sendBack(s);
		} else if (c < 0x80) {
			try {
				_fout.write(c);
				_fout.flush();
				e.consume();
			} catch (Exception ee) {
				// handle? probably means back-end is dead...
			}
		}
        }

        public void keyReleased(KeyEvent e) {
        }

	public void run() {
		// everything on 'fifo' is pasted text, throttle it.
		while (true) {
			String s = null;
			try {
				s = fifo.take();
			} catch (Exception ee) { }
			if (s == null) {
				continue;
			}
			try {
				for (byte b : s.getBytes()) {
					_fout.write(b);
					if (b == '\r') {
						Thread.sleep(cr_delay);
					} else {
						Thread.sleep(paste_delay);
					}
				}
				_fout.flush();
			} catch (Exception ee) {
			}
		}
	}
}
