// Copyright (c) 2016 Douglas Miller
import java.io.*;
import java.util.Map;
import java.util.HashMap;
import java.awt.event.*;
import javax.sound.sampled.*;
import java.util.Properties;
import java.util.concurrent.Semaphore;

public class H19Terminal implements AuxKeyListener, ActionListener, ResetListener {
	CrtScreen _crt;
	H19Keyboard _kbd;
	InputStream _fin;
	H19ANSIParser _ansi;
	int curs_x;
	int curs_y;
	int save_x;
	int save_y;
	String ansi;
	byte sw401;
	byte sw402;

	modes mode;
	boolean rev_video = false;
	boolean graphics = false;
	boolean insert = false;
	boolean ansi_mode = false;
	boolean wrap_mode = false;
	boolean need_wrap = false;

	// index into settings[]...
	static final int LINE25ON =    0;
	static final int KEYCLICKOFF = 1;
	static final int HOLDSCREEN =  2;
	static final int BLOCKCURSOR = 3;
	static final int CURSOROFF =   4;
	static final int KEYPADSHIFT = 5;
	static final int ALTKEYPAD =   6;
	static final int AUTOLFCR =    7;
	static final int AUTOCRLF =    8;

	boolean[] settings = new boolean[9];
	private Semaphore scrollWait;
	private int scroll;
	private boolean xoff;

	static final Map<Integer, String> altKeys;
	static {
		altKeys = new HashMap<Integer, String>();
		altKeys.put((int)'0', "\033?p");
		altKeys.put((int)'1', "\033?q");
		altKeys.put((int)'2', "\033?r");
		altKeys.put((int)'3', "\033?s");
		altKeys.put((int)'4', "\033?t");
		altKeys.put((int)'5', "\033?u");
		altKeys.put((int)'6', "\033?v");
		altKeys.put((int)'7', "\033?w");
		altKeys.put((int)'8', "\033?x");
		altKeys.put((int)'9', "\033?y");
		altKeys.put((int)'.', "\033?n");
		altKeys.put((int)'\n', "\033?M");
	}

	enum modes {
		NORMAL,
		ESC,
		SETCURS1,
		SETCURS2,
		BAUD,
		SETMODE,
		RESETMODE,
		ANSI,
	};

	Clip _beep;

	public H19Terminal(Properties props, CrtScreen crt, H19Keyboard kbd, InputStream fin) {
		mode = modes.NORMAL;
		ansi = new String();
		_crt = crt;
		_kbd = kbd;
		_fin = fin;
		_ansi = new H19ANSIParser(this);
		curs_x = 0;
		curs_y = 0;
		save_x = 0;
		save_y = 0;
		scrollWait = new Semaphore(0);
		xoff = false;
		_kbd.addAuxKeyListener(this);
		crt.setCursor(curs_x, curs_y);
		sw401 = (byte)(Integer.valueOf("10001100", 2) & 0x0ff);
		sw402 = (byte)(Integer.valueOf("00000110", 2) & 0x0ff);
		String s = props.getProperty("h19_s401");
		if (s != null) {
			sw401 = (byte)(Integer.valueOf(s, 2) & 0x0ff);
		}
		s = props.getProperty("h19_s402");
		if (s != null) {
			sw402 = (byte)(Integer.valueOf(s, 2) & 0x0ff);
		}
		int cps = -1;
		int crd = -1;
		s = props.getProperty("h19_paste_rate");
		if (s != null) {
			cps = Integer.valueOf(s);
		}
		s = props.getProperty("h19_paste_cr_wait");
		if (s != null) {
			crd = Integer.valueOf(s);
		}
		_kbd.setPasteRate(cps, crd);
		setupBeep(props);
		reset();
	}

	private void setupBeep(Properties props) {
		String s = props.getProperty("h19_beep");
		if (s == null) {
			s = "h19beep.wav";
		} else if (s.length() == 0) {
			_beep = null;
			return;
		}
		String beep_wav = s;
		try {
			AudioInputStream wav =
				AudioSystem.getAudioInputStream(
					new BufferedInputStream(
						this.getClass().getResourceAsStream(beep_wav)));
			AudioFormat format = wav.getFormat();
			DataLine.Info info = new DataLine.Info(Clip.class, format);
			_beep = (Clip)AudioSystem.getLine(info);
			_beep.open(wav);
			//_beep.setLoopPoints(0, loop);
		} catch (Exception e) {
			//e.printStackTrace();
			_beep = null;
			return;
		}
		int volume = 50;
		s = props.getProperty("h19_beep_volume");
		if (s != null) {
			volume = Integer.valueOf(s);
			if (volume < 0) volume = 0;
			if (volume > 100) volume = 100;
		}
		FloatControl vol = null;
		if (_beep.isControlSupported(FloatControl.Type.MASTER_GAIN)) {
			vol = (FloatControl)_beep.getControl(FloatControl.Type.MASTER_GAIN);
		} else if (_beep.isControlSupported(FloatControl.Type.VOLUME)) {
			vol = (FloatControl)_beep.getControl(FloatControl.Type.VOLUME);
		}
		if (vol != null) {
			float min = vol.getMinimum();
			float max = vol.getMaximum();
			float gain = (float)(min + ((max - min) * (volume / 100.0)));
			vol.setValue(gain);
		}
	}

	public void run() {
		while (true) {
			try {
				int c = _fin.read();
				if (c == -1) {
					break;
				}
				process_keychar(c);
			} catch(Exception ee) {
				break;
			}
		}
	}

	private void beep() {
		if (_beep != null) {
			_beep.setFramePosition(0);
			_beep.loop(0);
		}
	}

	private void reset() {
		curs_x = 0;
		curs_y = 0;
		_crt.setCursor(curs_x, curs_y);
		_crt.clearScreen();
		settings[BLOCKCURSOR] = ((sw402 & 0x01) != 0);
		settings[KEYCLICKOFF] = ((sw402 & 0x02) != 0);
		wrap_mode = ((sw402 & 0x04) != 0);
		settings[AUTOLFCR] = ((sw402 & 0x08) != 0);
		settings[AUTOCRLF] = ((sw402 & 0x10) != 0);
		ansi_mode = ((sw402 & 0x20) != 0);
		settings[KEYPADSHIFT] = ((sw402 & 0x40) != 0);
		// refresh50hz = ((sw402 & 0x80) != 0);
		settings[LINE25ON] = false;
		settings[HOLDSCREEN] = false;
		settings[CURSOROFF] = false;
		settings[ALTKEYPAD] = false;
		rev_video = false;
		graphics = false;
		insert = false;
		need_wrap = false;
		_crt.enableCursor(true);
		_crt.blockCursor(settings[BLOCKCURSOR]);
		mode = modes.NORMAL;
		// beep(); ???
	}

	private void set_modes(boolean on, int code) {
		code -= '1';
		if (code >= 9) {
			return;
		}
		settings[code] = on;
		if (code == CURSOROFF) {
			_crt.enableCursor(!on);
		} else if (code == BLOCKCURSOR) {
			_crt.blockCursor(on);
		}
		// NOTE: cannot receive this while actually HOLD
		if (code == HOLDSCREEN) {
			if (on) {
				scroll = 0;
				xoff = false; //?
			} else {
				doXon();
			}
		}
	}

	private void process_ansi(String seq) {
		if (seq.equals("?2h")) {
			ansi_mode = false;
		} else if (!_ansi.process_ansi(seq)) {
			System.err.format("Unimplemented ANSI sequence: %s\n", seq);
		}
	}

	// Note: these are ASCII codes, not KeyPress events...
	private void process_oth(int c) {
		if (c >= 0) { // dumb warnings...
			mode = modes.NORMAL;
		}
	}

	// Note: these are ASCII codes, not KeyPress events...
	private void process_ext(int c) {
		if (c == 'E') {
			curs_x = 0;
			curs_y = 0;
			_crt.setCursor(curs_x, curs_y);
			_crt.clearScreen();
		} else if (c == 'b') {
			_crt.clearBOP(curs_x, curs_y);
		} else if (c == 'J') {
			_crt.clearEOP(curs_x, curs_y);
		} else if (c == 'l') {
			_crt.clearLine(curs_y);
		} else if (c == 'o') {
			_crt.clearBOL(curs_x, curs_y);
		} else if (c == 'K') {
			_crt.clearEOL(curs_x, curs_y);
		} else if (c == 'L') {
			_crt.insertLine(curs_y);
		} else if (c == 'M') {
			_crt.deleteLine(curs_y);
		} else if (c == 'N') {
			_crt.deleteChar(curs_x, curs_y);
		} else if (c == '@') {
			insert = true;
		} else if (c == 'O') {
			insert = false;
		} else if (c == 'z') {
			reset();
		} else if (c == 'r') {
			mode = modes.BAUD;
			return;
		} else if (c == 'p') {
			rev_video = true;
		} else if (c == 'q') {
			rev_video = false;
		} else if (c == 'F') {
			graphics = true;
		} else if (c == 'G') {
			graphics = false;
		} else if (c == 'v') {
			wrap_mode = true;
		} else if (c == 'w') {
			wrap_mode = false;
		} else if (c == ']') {
			_kbd.sendBack(_crt.getScreen(24));
		} else if (c == '#') {
			_kbd.sendBack(_crt.getScreen(-1));
		} else if (c == 'x') {
			mode = modes.SETMODE;
			return;
		} else if (c == 'y') {
			mode = modes.RESETMODE;
			return;
		} else if (c == '<') {
			System.err.println("ANSI mode not supported");
			ansi_mode = true;
		} else {
			process_oth(c);
			return;
		}
		mode = modes.NORMAL;
	}

	// Note: these are ASCII codes, not KeyPress events...
	private void process_esc(int c) {
		if (ansi_mode) {
			mode = (c == '[' ? modes.ANSI : modes.NORMAL);
			return;
		}
		if (c == 'Z') {
			_kbd.sendBack(String.format("\033/K"));
			mode = modes.NORMAL;
			return;
		} else if (c == 'H') {
			curs_x = 0;
			curs_y = 0;
		} else if (c == 'C') {
			if (curs_x < 79) {
				++curs_x;
			}
		} else if (c == 'D') {
			if (curs_x > 0) {
				--curs_x;
			}
		} else if (c == 'B') {
			if (curs_y < 23 && curs_y < 24) {
				++curs_y;
			}
		} else if (c == 'A') {
			if (curs_y > 0 && curs_y < 24) {
				--curs_y;
			}
		} else if (c == 'I') {
			if (curs_y > 0) {
				if (curs_y < 24) {
					--curs_y;
				}
			} else {
				_crt.scrollDown();
				mode = modes.NORMAL;
				return;
			}
		} else if (c == 'n') {
			_kbd.sendBack(String.format("\033Y%c%c", (char)(curs_y + ' '), (char)(curs_x + ' ')));
		} else if (c == 'j') {
			save_x = curs_x;
			save_y = curs_y;
			mode = modes.NORMAL;
			return;
		} else if (c == 'k') {
			curs_x = save_x;
			curs_y = save_y;
		} else if (c == 'Y') {
			mode = modes.SETCURS1;
			return;
		} else if (c == 't') {
			set_modes(true, '6');
		} else if (c == 'u') {
			set_modes(false, '6');
		} else if (c == '=') {
			set_modes(true, '7');
		} else if (c == '>') {
			set_modes(false, '7');
		} else if (c == '\\') {
			set_modes(false, '3');
		} else if (c == '[') {
			if (ansi_mode) {
				mode = modes.ANSI;
			} else {
				set_modes(true, '3');
				mode = modes.NORMAL;
			}
			return;
		} else {
			process_ext(c);
			return;
		}
		_crt.setCursor(curs_x, curs_y);
		need_wrap = false;
		mode = modes.NORMAL;
	}

	public void set_curs_xy(int x, int y) {
		set_curs_y(y);
		set_curs_x(x);
		_crt.setCursor(curs_x, curs_y);
	}

	private void set_curs_y(int c) {
		if (c < 25 && (c < 24 || settings[LINE25ON])) {
			curs_y = c;
		}
	}

	private void set_curs_x(int c) {
		if (c >= 80) c = 79;
		curs_x = c;
	}

	private void doLinefeed() {
		if (curs_y < 23) {
			++curs_y;
		} else if (curs_y >= 24) {
			// linefeed ignored on 25th line
		} else if (settings[HOLDSCREEN]) {
			--scroll;
			if (scroll <= 0) {
				// According to H19 source code, we should
				// hang here until SCROLL pressed. Since
				// we are running in our own thread, we
				// could theoretically do that. We need
				// to actually perform the LF once SCROLL
				// is pressed...
				// TODO: implement 128 byte input buffer,
				// with XON/XOFF thresholds.
				doXoff();
				scrollWait.drainPermits();
				try {
					scrollWait.acquire();
				} catch (Exception ee) {}
				doXon();
			}
			_crt.scrollUp();
		} else {
			_crt.scrollUp();
		}
	}

	// Note: these are ASCII codes, not KeyPress events...
	public void process_keychar(int c) {
		c &= 0x7f;
		if (c == 0x1b) {
			mode = modes.ESC;
			return;
		}
		if (mode == modes.ESC) {
			process_esc(c);
			return;
		} else if (mode == modes.SETCURS1) {
			set_curs_y(c - ' ');
			mode = modes.SETCURS2;
			return;
		} else if (mode == modes.SETCURS2) {
			set_curs_x(c - ' ');
			_crt.setCursor(curs_x, curs_y);
			mode = modes.NORMAL;
			return;
		} else if (mode == modes.BAUD) {
			mode = modes.NORMAL;
			return;
		} else if (mode == modes.SETMODE) {
			set_modes(true, c);
			mode = modes.NORMAL;
			return;
		} else if (mode == modes.RESETMODE) {
			set_modes(false, c);
			mode = modes.NORMAL;
			return;
		} else if (mode == modes.ANSI) {
			ansi += (char)c;
			// stupid bugs...
			// why are 'H' and 'f' false but all others are true?!
			// even "if (c >= 0x41) {" does not fix it.
//System.err.format("ANSI %08x\n", c);
			if (Character.isLetter((char)c)) {
				process_ansi(ansi);
				ansi = new String();
				mode = modes.NORMAL;
			}
			return;
		} else if (c == 0x0a) {
			if (settings[AUTOCRLF]) {
				need_wrap = false;
				curs_x = 0;
			}
			doLinefeed();
		} else if (c == 0x0d) {
			need_wrap = false;
			curs_x = 0;
			if (settings[AUTOLFCR]) {
				doLinefeed();
			}
		} else if (c == 0x09) {
			if (curs_x < 72) {
				curs_x = curs_x + (8 - (curs_x & 7));
			} else {
				curs_x = 79;
			}
		} else if (c == 0x08) {
			need_wrap = false;
			if (curs_x > 0) {
				--curs_x;
			}
		} else if (c == 0x07) {
			beep();
		} else if (c < ' ') {
			// TODO: handle other Ctrl chars?
			return;
		} else {
			// might need to translate character based on rev-vid, etc.
			c = _crt.specialChar(c, rev_video, graphics);
			if (wrap_mode && curs_x == 79 && need_wrap) {
				curs_x = 0;
				// according to H19 source code, this
				// does not honor HOLD SCREEN mode.
				if (curs_y >= 23) {
					if (curs_y < 24) {
						_crt.scrollUp();
					}
				} else {
					++curs_y;
				}
			}
			if (insert) {
				_crt.insertChar((char)c, curs_x, curs_y);
			} else {
				_crt.putChar((char)c, curs_x, curs_y);
			}
			need_wrap = false;
			if (curs_x < 79) {
				++curs_x;
			} else if (wrap_mode) {
				need_wrap = true;
			}
		}
		_crt.setCursor(curs_x, curs_y);
	}

	public void actionPerformed(ActionEvent e) {}

	public void resetPerformed() {
		reset();
	}

	private void doXoff() {
		if (!xoff) {
			xoff = true;
			_kbd.sendBack("\023"); // XOFF, ^S
		}
	}

	private void doXon() {
		if (xoff) {
			xoff = false;
			_kbd.sendBack("\021"); // XON, ^Q
		}
	}

	private void doInsertKey() {
		if (insert) {
			_kbd.sendBack("\033O"); // exit insert mode
		} else {
			_kbd.sendBack("\033@"); // enter insert mode
		}
	}

	private void doScrollKey(boolean shifted) {
		if (!settings[HOLDSCREEN]) {
			return;
		}
		if (scrollWait.availablePermits() > 0) {
			return;
		}
		if (shifted) {
			scroll = 24;
		} else {
			scroll = 1;
		}
		scrollWait.release();
	}

	public void auxKeyPressed(KeyEvent e) {
		int c = e.getKeyChar();
		int k = e.getKeyCode();
		int l = e.getKeyLocation();
		int m = e.getModifiers();

		if (k == KeyEvent.VK_INSERT) { // must be unshifted...
			doInsertKey();
		} else if (k == KeyEvent.VK_SCROLL_LOCK) {
			doScrollKey((m & InputEvent.SHIFT_MASK) != 0);
		} else if (l == KeyEvent.KEY_LOCATION_NUMPAD) {
			if (!settings[ALTKEYPAD]) {
				_kbd.sendBack(new String(new byte[]{(byte)c}));
			} else if (altKeys.containsKey(c)) {
				_kbd.sendBack(altKeys.get(c));
			}
		}
	}
}
