// Copyright (c) 2016 Douglas Miller
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.Properties;

public class h19_fe {
	final String ident = "$Id$";

	private static JFrame front_end;

	public static void main(String[] args) {
		java.io.OutputStream fout = null;
		java.io.InputStream fin = null;

		java.net.URL url = h19_fe.class.getResource("h19_fe.class");
		String jar = "jar:" + url.getFile().replace("h19_fe.class", "");

		boolean loopback = (args.length > 0 && args[0].compareTo("loopback") == 0);
		Properties props = new H19Properties("19", args);
		String f = "H19a.ttf";
		float fz = 20f;
		Color fc = Color.green;
		String s = props.getProperty("h19_font");
		if (s != null) {
			f = s;
		}
		s = props.getProperty("h19_font_size");
		if (s != null) {
			fz = Float.valueOf(s);
		}
		s = props.getProperty("h19_font_color");
		if (s != null) {
			fc = new Color(Integer.valueOf(s, 16));
		}

		CrtScreen screen = new CrtScreen(fz);
		Font font = null;
		if (f.endsWith(".ttf")) {
			try {
				File ff = new File(f);
				java.io.InputStream ttf;
				if (ff.exists()) {
					ttf = new FileInputStream(ff);
				} else {
					ttf = h19_fe.class.getResourceAsStream(f);
				}
				if (ttf != null) {
					font = Font.createFont(Font.TRUETYPE_FONT, ttf);
					font = font.deriveFont(fz);
				}
			} catch (Exception ee) {
				font = null;
			}
		} else {
			font = new Font(f, Font.PLAIN, (int)fz);
			screen.setISO();
		}
		if (font == null) {
			System.err.println("Missing font \"" +
					f + "\", using default");
			font = new Font("Monospaced", Font.PLAIN, 10);
		}
		screen.setFont(font);
		screen.setForeground(fc);

		front_end = new JFrame("Virtual H19 Terminal");
		front_end.getContentPane().setName("H19 Emulator");
		front_end.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		front_end.getContentPane().setBackground(new Color(100, 100, 100));
		// This allows TAB to be sent
		front_end.setFocusTraversalKeysEnabled(false);
		s = props.getProperty("h19_nameplate");
		if (s == null) {
			s =	"<HTML>&nbsp;" +
				"Heathkit H19" +
				"&nbsp;</HTML>";
		}
		if (s.equals("none")) {
			front_end.add(screen);
		} else {
			s = s.replaceAll("jar:", jar);
			JPanel front = new HxxFrontSide(front_end, screen, props, s);
			front_end.add(front);
		}

		if (loopback) {
			fin = new PipedInputStream();
			try {
				fout = new PipedOutputStream((PipedInputStream)fin);
			} catch(Exception ee) {
				System.err.println("Unable to create pipe!");
				System.exit(1);
			}
		} else {
			fin = System.in;
			fout = System.out;
		}

		H19Keyboard kbd = new H19Keyboard(fout, loopback);
		front_end.addKeyListener(kbd);
		screen.setPasteListener(kbd);
		H19Terminal term = new H19Terminal(props, screen, kbd, fin);
		kbd.addResetListener(term);

		front_end.pack();
		front_end.setVisible(true);

		term.run();
	}
}
