// Copyright (c) 2016 Douglas Miller
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.Properties;

class FEexit extends Thread {
	private Process _be = null;
	private OutputStream _fout;

	public FEexit(Process be, OutputStream fout) {
		_be = be;
		_fout = fout;
	}
	public void run() {
		if (_be != null) {
			byte[] quit = new String("quit\n").getBytes();
			try {
				_fout.write(quit);
				_fout.flush();
			} catch (Exception ee) {
				_be.destroy();
			}
			try {
				_be.waitFor();
			} catch (InterruptedException ee) {
			}
		}
	}
}

public class h89_fe {
	final String ident = "$Id$";

	private static JFrame front_end;

	public static void main(String[] args) {
		java.io.OutputStream fout = null;
		java.io.InputStream fin = null;
		//java.io.BufferedReader ferr = null;
		String cmd = "./v89 -q -g proxy";

		java.net.URL url = h89_jni.class.getResource("h89_fe.class");
		String jar = "jar:" + url.getFile().replace("h89_fe.class", "");

		Properties props = new H19Properties("89", args);
		String s = props.getProperty("h89_cmd_line");
		if (s != null) {
			cmd = s;
		}

		String f = "H19a.ttf";
		float fz = 20f;
		Color fc = Color.green;
		s = props.getProperty("h19_font");
		if (s != null) {
			f = s;
		}
		s = props.getProperty("h19_font_size");
		if (s != null) {
			fz = Float.valueOf(s);
		}
		s = props.getProperty("h19_font_color");
		if (s != null) {
			fc = new Color(Integer.valueOf(s, 16));
		}

		CrtScreen screen = new CrtScreen(fz);
		Font font = null;
		if (f.endsWith(".ttf")) {
			try {
				File ff = new File(f);
				java.io.InputStream ttf;
				if (ff.exists()) {
					ttf = new FileInputStream(ff);
				} else {
					ttf = h89_fe.class.getResourceAsStream(f);
				}
				if (ttf != null) {
					font = Font.createFont(Font.TRUETYPE_FONT, ttf);
					font = font.deriveFont(fz);
				}
			} catch (Exception ee) {
				font = null;
			}
		} else {
			font = new Font(f, Font.PLAIN, (int)fz);
			screen.setISO();
		}
		if (font == null) {
			System.err.println("Missing font \"" +
					f + "\", using default");
			font = new Font("Monospaced", Font.PLAIN, 10);
		}
		screen.setFont(font);
		screen.setForeground(fc);

		front_end = new JFrame("Virtual H89 Computer");
		front_end.getContentPane().setName("H89 Emulator");
		front_end.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		front_end.getContentPane().setBackground(new Color(100, 100, 100));
		// This allows TAB to be sent
		front_end.setFocusTraversalKeysEnabled(false);
		s = props.getProperty("h89_nameplate");
		if (s == null) {
			s =	"<HTML>&nbsp;" +
				"Heathkit H89" +
				"&nbsp;</HTML>";
		}
		if (s.equals("none")) {
			front_end.add(screen);
		} else {
			s = s.replaceAll("jar:", jar);
			JPanel front = new HxxFrontSide(front_end, screen, props, s);
			front_end.add(front);
		}

		H89Operator op = new H89Operator(front_end, props, screen, null);

		OutputStream f0out = null;
		InputStream f0in = null;
		OutputStream f1out = null;
		InputStream f1in = null;
		f0in = new PipedInputStream();
		try {
			f0out = new PipedOutputStream((PipedInputStream)f0in);
		} catch(Exception ee) {
			System.err.println("Unable to create pipe 0!");
			System.exit(1);
		}
		f1in = new PipedInputStream();
		try {
			f1out = new PipedOutputStream((PipedInputStream)f1in);
		} catch(Exception ee) {
			System.err.println("Unable to create pipe 1!");
			System.exit(1);
		}

		try {
			Process _be = null;
			_be = Runtime.getRuntime().exec(cmd);
			fout = _be.getOutputStream();
			fin = _be.getInputStream();
			//ferr = new BufferedReader(new InputStreamReader(_be.getErrorStream()));
			Runtime.getRuntime().addShutdownHook(new FEexit(_be, fout));
		} catch (IOException ee) {
			System.err.println("Unable to exec back-end!");
			System.exit(1);
		}
		H89ResponseTee rt = new H89ResponseTee(fin, f1out);
		H89CommandTee ct = new H89CommandTee(f0in, fout, rt);
		op.setCommander(ct);

		H19Keyboard kbd = new H19Keyboard(f0out, false);
		front_end.addKeyListener(kbd);
		screen.setPasteListener(kbd);
		H19Terminal term = new H19Terminal(props, screen, kbd, f1in);
		kbd.addResetListener(term);
		kbd.addResetListener(op);

		front_end.pack();
		front_end.setVisible(true);

		term.run();
	}
}
