// Copyright (c) 2016 Douglas Miller
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.Properties;
import java.lang.reflect.Field;

public class h89_jni {
	// A bit brute-force, but it works. However, it depends on the name of
	// an internal field, so it could break in later versions of Java or
	// on other platforms. This is all necessary because the ClassLoader
	// has already initialized "sys_paths" by the time this static block
	// runs. Worst case, use "-Djava.library.path=." on commandline.
	static {
		System.setProperty("java.library.path", System.getProperty("user.dir"));
		try {
			Field fieldSysPath = ClassLoader.class.getDeclaredField("sys_paths");
			fieldSysPath.setAccessible(true);
			fieldSysPath.set(null, null);
		} catch (Exception ee) {}
	}

	private static JFrame front_end;

	public static void main(String[] args) {
		String cmd = "./v89 -q -g proxy";

		java.net.URL url = h89_jni.class.getResource("h89_jni.class");
		String jar = "jar:" + url.getFile().replace("h89_jni.class", "");
		//System.err.println(jar);
		// jar:file:/path/to/jar/H89jni.jar!/

		Properties props = new H19Properties("89", args);
		String s = props.getProperty("h89_cmd_line");
		if (s != null) {
			cmd = s;
		}

		String f = "H19a.ttf";
		float fz = 20f;
		Color fc = Color.green;
		s = props.getProperty("h19_font");
		if (s != null) {
			f = s;
		}
		s = props.getProperty("h19_font_size");
		if (s != null) {
			fz = Float.valueOf(s);
		}
		s = props.getProperty("h19_font_color");
		if (s != null) {
			fc = new Color(Integer.valueOf(s, 16));
		}

		CrtScreen screen = new CrtScreen(fz);
		Font font = null;
		if (f.endsWith(".ttf")) {
			try {
				File ff = new File(f);
				java.io.InputStream ttf;
				if (ff.exists()) {
					ttf = new FileInputStream(ff);
				} else {
					ttf = h19_fe.class.getResourceAsStream(f);
				}
				if (ttf != null) {
					font = Font.createFont(Font.TRUETYPE_FONT, ttf);
					font = font.deriveFont(fz);
				}
			} catch (Exception ee) {
				font = null;
			}
		} else {
			font = new Font(f, Font.PLAIN, (int)fz);
			screen.setISO();
		}
		if (font == null) {
			System.err.println("Missing font \"" +
					f + "\", using default");
			font = new Font("Monospaced", Font.PLAIN, 10);
		}
		screen.setFont(font);
		screen.setForeground(fc);

		front_end = new JFrame("Virtual H89 Computer");
		front_end.getContentPane().setName("H89 Emulator");
		front_end.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		front_end.getContentPane().setBackground(new Color(100, 100, 100));
		// This allows TAB to be sent
		front_end.setFocusTraversalKeysEnabled(false);
		s = props.getProperty("h89_nameplate");
		if (s == null) {
			s =	"<HTML>&nbsp;" +
				"Heathkit H89" +
				"&nbsp;</HTML>";
		}
		if (s.equals("none")) {
			front_end.add(screen);
		} else {
			s = s.replaceAll("jar:", jar);
			JPanel front = new HxxFrontSide(front_end, screen, props, s);
			front_end.add(front);
		}

		V89JNI jni = new V89JNI();
		jni.startV89(cmd.split("\\s+"));

		H89Operator op = new H89Operator(front_end, props, screen, null);
		H89CommandJni ct = new H89CommandJni(jni);
		op.setCommander(ct);

		H19Keyboard kbd = new H19Keyboard(ct, false);
		front_end.addKeyListener(kbd);
		screen.setPasteListener(kbd);
		H19Terminal term = new H19Terminal(props, screen, kbd, jni);
		kbd.addResetListener(term);
		kbd.addResetListener(op);

		front_end.pack();
		front_end.setVisible(true);

		term.run();
	}
}
